# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import sqliteog_pb2 as sqliteog__pb2


class SqliteOGStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Query = channel.unary_unary(
                '/SqliteOG/Query',
                request_serializer=sqliteog__pb2.Statement.SerializeToString,
                response_deserializer=sqliteog__pb2.QueryResult.FromString,
                )
        self.Execute = channel.unary_unary(
                '/SqliteOG/Execute',
                request_serializer=sqliteog__pb2.Statement.SerializeToString,
                response_deserializer=sqliteog__pb2.ExecuteResult.FromString,
                )
        self.ExecuteOrQuery = channel.unary_unary(
                '/SqliteOG/ExecuteOrQuery',
                request_serializer=sqliteog__pb2.Statement.SerializeToString,
                response_deserializer=sqliteog__pb2.ExecuteOrQueryResult.FromString,
                )
        self.Callback = channel.stream_stream(
                '/SqliteOG/Callback',
                request_serializer=sqliteog__pb2.InvocationResult.SerializeToString,
                response_deserializer=sqliteog__pb2.Invoke.FromString,
                )
        self.Connection = channel.unary_unary(
                '/SqliteOG/Connection',
                request_serializer=sqliteog__pb2.ConnectionRequest.SerializeToString,
                response_deserializer=sqliteog__pb2.ConnectionId.FromString,
                )
        self.Close = channel.unary_unary(
                '/SqliteOG/Close',
                request_serializer=sqliteog__pb2.ConnectionId.SerializeToString,
                response_deserializer=sqliteog__pb2.Empty.FromString,
                )
        self.IsValid = channel.unary_unary(
                '/SqliteOG/IsValid',
                request_serializer=sqliteog__pb2.ConnectionId.SerializeToString,
                response_deserializer=sqliteog__pb2.Empty.FromString,
                )
        self.Ping = channel.unary_unary(
                '/SqliteOG/Ping',
                request_serializer=sqliteog__pb2.Empty.SerializeToString,
                response_deserializer=sqliteog__pb2.Empty.FromString,
                )
        self.ResetSession = channel.unary_unary(
                '/SqliteOG/ResetSession',
                request_serializer=sqliteog__pb2.ConnectionId.SerializeToString,
                response_deserializer=sqliteog__pb2.ConnectionId.FromString,
                )


class SqliteOGServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Query(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Execute(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ExecuteOrQuery(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Callback(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Connection(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Close(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def IsValid(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Ping(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ResetSession(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_SqliteOGServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Query': grpc.unary_unary_rpc_method_handler(
                    servicer.Query,
                    request_deserializer=sqliteog__pb2.Statement.FromString,
                    response_serializer=sqliteog__pb2.QueryResult.SerializeToString,
            ),
            'Execute': grpc.unary_unary_rpc_method_handler(
                    servicer.Execute,
                    request_deserializer=sqliteog__pb2.Statement.FromString,
                    response_serializer=sqliteog__pb2.ExecuteResult.SerializeToString,
            ),
            'ExecuteOrQuery': grpc.unary_unary_rpc_method_handler(
                    servicer.ExecuteOrQuery,
                    request_deserializer=sqliteog__pb2.Statement.FromString,
                    response_serializer=sqliteog__pb2.ExecuteOrQueryResult.SerializeToString,
            ),
            'Callback': grpc.stream_stream_rpc_method_handler(
                    servicer.Callback,
                    request_deserializer=sqliteog__pb2.InvocationResult.FromString,
                    response_serializer=sqliteog__pb2.Invoke.SerializeToString,
            ),
            'Connection': grpc.unary_unary_rpc_method_handler(
                    servicer.Connection,
                    request_deserializer=sqliteog__pb2.ConnectionRequest.FromString,
                    response_serializer=sqliteog__pb2.ConnectionId.SerializeToString,
            ),
            'Close': grpc.unary_unary_rpc_method_handler(
                    servicer.Close,
                    request_deserializer=sqliteog__pb2.ConnectionId.FromString,
                    response_serializer=sqliteog__pb2.Empty.SerializeToString,
            ),
            'IsValid': grpc.unary_unary_rpc_method_handler(
                    servicer.IsValid,
                    request_deserializer=sqliteog__pb2.ConnectionId.FromString,
                    response_serializer=sqliteog__pb2.Empty.SerializeToString,
            ),
            'Ping': grpc.unary_unary_rpc_method_handler(
                    servicer.Ping,
                    request_deserializer=sqliteog__pb2.Empty.FromString,
                    response_serializer=sqliteog__pb2.Empty.SerializeToString,
            ),
            'ResetSession': grpc.unary_unary_rpc_method_handler(
                    servicer.ResetSession,
                    request_deserializer=sqliteog__pb2.ConnectionId.FromString,
                    response_serializer=sqliteog__pb2.ConnectionId.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'SqliteOG', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class SqliteOG(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Query(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/SqliteOG/Query',
            sqliteog__pb2.Statement.SerializeToString,
            sqliteog__pb2.QueryResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Execute(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/SqliteOG/Execute',
            sqliteog__pb2.Statement.SerializeToString,
            sqliteog__pb2.ExecuteResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ExecuteOrQuery(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/SqliteOG/ExecuteOrQuery',
            sqliteog__pb2.Statement.SerializeToString,
            sqliteog__pb2.ExecuteOrQueryResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Callback(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/SqliteOG/Callback',
            sqliteog__pb2.InvocationResult.SerializeToString,
            sqliteog__pb2.Invoke.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Connection(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/SqliteOG/Connection',
            sqliteog__pb2.ConnectionRequest.SerializeToString,
            sqliteog__pb2.ConnectionId.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Close(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/SqliteOG/Close',
            sqliteog__pb2.ConnectionId.SerializeToString,
            sqliteog__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def IsValid(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/SqliteOG/IsValid',
            sqliteog__pb2.ConnectionId.SerializeToString,
            sqliteog__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Ping(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/SqliteOG/Ping',
            sqliteog__pb2.Empty.SerializeToString,
            sqliteog__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ResetSession(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/SqliteOG/ResetSession',
            sqliteog__pb2.ConnectionId.SerializeToString,
            sqliteog__pb2.ConnectionId.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: sqliteog.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0esqliteog.proto\"\x07\n\x05\x45mpty\"\x1a\n\x0c\x43onnectionId\x12\n\n\x02id\x18\x01 \x01(\t\"L\n\x11\x43onnectionRequest\x12\x0f\n\x07\x64\x62_name\x18\x01 \x01(\t\x12\x11\n\tfunctions\x18\x02 \x03(\t\x12\x13\n\x0b\x61ggregators\x18\x03 \x03(\t\"3\n\x10InvocationResult\x12\x0f\n\x07initial\x18\x01 \x01(\x08\x12\x0e\n\x06result\x18\x02 \x03(\t\",\n\x06Invoke\x12\x14\n\x0c\x66unctionName\x18\x01 \x01(\t\x12\x0c\n\x04\x61rgs\x18\x02 \x03(\t\"b\n\x14\x45xecuteOrQueryResult\x12\"\n\x0cquery_result\x18\x01 \x01(\x0b\x32\x0c.QueryResult\x12&\n\x0e\x65xecute_result\x18\x02 \x01(\x0b\x32\x0e.ExecuteResult\"8\n\tStatement\x12\x0b\n\x03sql\x18\x01 \x01(\t\x12\x0e\n\x06params\x18\x02 \x03(\t\x12\x0e\n\x06\x63nx_id\x18\x03 \x01(\t\"\x15\n\x03Row\x12\x0e\n\x06\x66ields\x18\x01 \x03(\t\"G\n\x0bQueryResult\x12\x0f\n\x07\x63olumns\x18\x01 \x03(\t\x12\x13\n\x0b\x63olumnTypes\x18\x02 \x03(\t\x12\x12\n\x04rows\x18\x03 \x03(\x0b\x32\x04.Row\";\n\rExecuteResult\x12\x14\n\x0clastInsertId\x18\x01 \x01(\x03\x12\x14\n\x0c\x61\x66\x66\x65\x63tedRows\x18\x02 \x01(\x03\x32\x80\x03\n\x08SqliteOG\x12#\n\x05Query\x12\n.Statement\x1a\x0c.QueryResult\"\x00\x12\'\n\x07\x45xecute\x12\n.Statement\x1a\x0e.ExecuteResult\"\x00\x12\x35\n\x0e\x45xecuteOrQuery\x12\n.Statement\x1a\x15.ExecuteOrQueryResult\"\x00\x12,\n\x08\x43\x61llback\x12\x11.InvocationResult\x1a\x07.Invoke\"\x00(\x01\x30\x01\x12\x31\n\nConnection\x12\x12.ConnectionRequest\x1a\r.ConnectionId\"\x00\x12 \n\x05\x43lose\x12\r.ConnectionId\x1a\x06.Empty\"\x00\x12\"\n\x07IsValid\x12\r.ConnectionId\x1a\x06.Empty\"\x00\x12\x18\n\x04Ping\x12\x06.Empty\x1a\x06.Empty\"\x00\x12.\n\x0cResetSession\x12\r.ConnectionId\x1a\r.ConnectionId\"\x00\x42 Z\x1egithub.com/aousomran/sqlite-ogb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'sqliteog_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'Z\036github.com/aousomran/sqlite-og'
  _globals['_EMPTY']._serialized_start=18
  _globals['_EMPTY']._serialized_end=25
  _globals['_CONNECTIONID']._serialized_start=27
  _globals['_CONNECTIONID']._serialized_end=53
  _globals['_CONNECTIONREQUEST']._serialized_start=55
  _globals['_CONNECTIONREQUEST']._serialized_end=131
  _globals['_INVOCATIONRESULT']._serialized_start=133
  _globals['_INVOCATIONRESULT']._serialized_end=184
  _globals['_INVOKE']._serialized_start=186
  _globals['_INVOKE']._serialized_end=230
  _globals['_EXECUTEORQUERYRESULT']._serialized_start=232
  _globals['_EXECUTEORQUERYRESULT']._serialized_end=330
  _globals['_STATEMENT']._serialized_start=332
  _globals['_STATEMENT']._serialized_end=388
  _globals['_ROW']._serialized_start=390
  _globals['_ROW']._serialized_end=411
  _globals['_QUERYRESULT']._serialized_start=413
  _globals['_QUERYRESULT']._serialized_end=484
  _globals['_EXECUTERESULT']._serialized_start=486
  _globals['_EXECUTERESULT']._serialized_end=545
  _globals['_SQLITEOG']._serialized_start=548
  _globals['_SQLITEOG']._serialized_end=932
# @@protoc_insertion_point(module_scope)



import grpc
from .gen.sqliteog_pb2_grpc import SqliteOGStub
from .gen.sqliteog_pb2 import *
import datetime
from ast import literal_eval
from queue import Queue
from threading import Thread
from .errors import *

# Globals
paramstyle = "qmark"
threadsafety = 1
apilevel = "2.0"
PARSE_COLNAMES = 2
PARSE_DECLTYPES = 1
sqlite_version = '3.31.1'
version = '0.0.3'
version_info = tuple([int(x) for x in version.split(".")])
sqlite_version_info = tuple([int(x) for x in sqlite_version.split(".")])

Date = datetime.date
Time = datetime.time
Timestamp = datetime.datetime


# Connection
def connect(host, port, dbname):
    return Connection(host, port, dbname)


class Connection:
    channel = None
    grpc_stub = None
    _function_register = dict()
    initialized = False
    _cnx_id = ""
    _db_name = ""

    @property
    def cnx_id(self):
        return self._cnx_id

    @property
    def function_register(self):
        return self._function_register

    def __init__(self, host, port, dbname):
        self._db_name = dbname
        self.channel = grpc.insecure_channel(f"{host}:{port}")
        self.grpc_stub = SqliteOGStub(self.channel)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        # fire a call to close our remote db connection, before closing the grpc channel
        self.grpc_stub.Close(ConnectionId(id=self.cnx_id))
        self.channel.close()

    def initialize_remote_connection(self):
        resp: ConnectionId = self.grpc_stub.Connection(
            ConnectionRequest(db_name=self._db_name, functions=self._function_register.keys())
        )
        self._cnx_id = resp.id
        self.initialized = True

    def commit(self):
        pass

    def rollback(self):
        pass

    def cursor(self, factory=None):
        # let's initialize our connection before the first cursor call
        # we're waiting this long to make sure that all custom functions
        # have been registered
        if not self.initialized:
            self.initialize_remote_connection()
        if factory:
            return factory(self)
        return Cursor(self)

    def execute(self, sql, params=None):
        """
        for some reason the sqlite dbapi supports execute directly on the connection
        django also uses this when registering functions
        """
        cursor = self.cursor()
        cursor.execute(sql, params)
        return cursor

    def create_function(self, *args, **kwargs):
        name = args[0]
        num_args = args[1]
        func = args[2]
        self._function_register[name] = func

    def create_aggregate(self, *args, **kwargs):
        pass


class Cursor:
    grpc_stub = None
    connection = None
    description = [None for i in range(7)]
    rowcount = 0
    arraysize = 0

    # the following _vars represent a "buffer" of sorts
    # values would be set to the results from the latest .execute() call
    # this is because callers may not fetch all results at once, so we'd
    # need to keep results until "buffer" is emptied...
    _statement = None
    _rows = []
    _columns = []
    _last_insert_id = -1
    _affected_rows = -1

    _callback_queue = None
    _callback_thread = None
    _callback_invocations = None
    _sentinel = object()

    @property
    def lastrowid(self):
        return self._last_insert_id

    def __init__(self, connection):
        self.connection = connection
        self.grpc_stub = self.connection.grpc_stub
        self._init_callback()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _process_invocations(self):
        try:
            # blocks until we get an invocation from the generator
            for invoke in self._callback_invocations:
                function = self.connection.function_register.get(invoke.functionName)
                if not function:
                    raise Error(f"function {invoke.functionName} not registered")
                result = function(*invoke.args)
                self._callback_queue.put(result)
        except grpc.RpcError as e:
            # we're cancelling the rpc on cursor.close()
            if e.code() != grpc.StatusCode.CANCELLED:
                raise e

    def _invocation_results_iterator(self):
        # block until we get a result from queue, the sentinel value should end the iterator
        for i in iter(self._callback_queue.get, self._sentinel):
            if i != self._sentinel:
                yield InvocationResult(result=[f"{i}"])

    def _init_callback(self):
        self._callback_queue = Queue(maxsize=0)
        # add the connection id to the request metadata
        metadata = [('cnx_id', self.connection.cnx_id)]
        self._callback_invocations = self.grpc_stub.Callback(self._invocation_results_iterator(), metadata=metadata)
        self._callback_thread = Thread(target=self._process_invocations)
        self._callback_thread.start()

    def close(self, *args, **kwargs):
        """ Closes the cursor. """
        self._callback_invocations.cancel()
        self._callback_thread.join(0)

    def _prepare_params(self, params):
        if params and len(params) > 0:
            # at this time using string types for all request & result params
            return (str(x) for x in params)
        return None

    def _cast_row_fields(self, fields):
        result = []
        for f in fields:
            try:
                f = literal_eval(f)
            except:
                pass
            result.append(f)
        return tuple(result)

    def _results_to_buffer(self, results: ExecuteOrQueryResult):
        self._columns = results.query_result.columns
        self._last_insert_id = results.execute_result.lastInsertId
        self._affected_rows = results.execute_result.affectedRows
        self._rows = [self._cast_row_fields(row.fields) for row in results.query_result.rows]

    def reset_buffer(self):
        self._rows = []
        self._columns = []
        self._last_insert_id = -1
        self._affected_rows = -1

    def execute(self, sql, params=None):
        """
        Executes remote query & put the result in the buffer
        """
        # reset buffer before we start executing, so that if there's a failure
        # we won't have any stale data
        self.reset_buffer()
        # we require all string params for now
        params = self._prepare_params(params)
        self._statement = Statement(sql=sql, params=params, cnx_id=self.connection.cnx_id)
        results: ExecuteOrQueryResult = self.grpc_stub.ExecuteOrQuery(self._statement)
        self._results_to_buffer(results)
        return self

    def executemany(self, *args, **kwargs):
        """ Repeatedly executes a SQL statement. """
        pass

    def executescript(self, *args, **kwargs):
        """ Executes a multiple SQL statements at once. Non-standard. """
        pass

    def fetchall(self):
        """ Fetches all (remaining) rows from the buffer """
        self._rows, fetched = [], self._rows
        return fetched

    def fetchmany(self, size):
        """ Fetches n=size rows and removes them from the buffer """
        fetched = []
        if size < len(self._rows):
            self._rows, fetched = self._rows[size:], self._rows[:size]
        else:
            self._rows, fetched = [], self._rows
        return fetched

    def fetchone(self, *args, **kwargs):
        """ Fetches one row from the buffer """
        rows = self.fetchmany(1)
        # avoid index out of range
        if len(rows) < 1:
            result = rows
            return result
        result = rows[0]
        return result

    def setinputsizes(self, *args, **kwargs):
        """ Required by DB-API """
        pass

    def setoutputsize(self, *args, **kwargs):
        """ Required by DB-API """
        pass



from pysqliteog import dbapi2

table_sql = """
		CREATE TABLE IF NOT EXISTS example_table (
			id INTEGER PRIMARY KEY,
			name TEXT,
			age INTEGER,
			height REAL,
			is_student INTEGER
		);
"""


def say_hello(x):
    return f'hello {x}'


def square(x):
    return int(x) * int(x)


def test_connection():
    db = dbapi2.connect('localhost', '9090', 'testpy.db')
    db.create_function('say_hello', 1, say_hello)
    db.create_function('square', 1, square)
    with db.cursor() as cur:
        cur.execute(table_sql)

    with db.cursor() as cur:
        cur.execute("select square(?)", params=[2])
        print(cur.fetchone())

    with db.cursor() as cur:
        cur.execute("select say_hello(?)", params=['ao'])
        print(cur.fetchone())

    with db.cursor() as cur:
        sql = """INSERT INTO example_table (name, age, height, is_student) values(?,?,?,?)"""
        cur.execute(sql, params=["mike jones", 40, 192, True])
        cur.execute("SELECT * FROM example_table")
        print(cur.fetchall())
    db.close()


if __name__ == "__main__":
    test_connection()
